// ------------------------------------------------------------
//  "The Star" by Neil Parris, 2020
// ------------------------------------------------------------
// Simple LED animations driving 10x digital outputs connected to LEDs.
// Designed for use with a custom star-shaped PCB which has 10x LEDs on each size.
// Uses a simple software PWM routine.
// 
// Requires FastLED library for helper functions - https://github.com/FastLED/FastLED
//

/* The MIT License (MIT)

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

// include the FastLED library as it has some fast helper functions like sin8();
#include "FastLED.h"

// create an array of the *arduino* pin names connected to the LEDS, this is
// the same mapping as an Arduino Nano
// A2 = LED1 (physical MCU pin 25)
// A3 = LED2 (physcial MCU pin 26)
// etc.
// D4 = LED10

const byte ledpins[] = {A2, A3, A1, A0, 9, 10, 6, 5, 3, 4};

// physical push button pins (not yet used in the code)
#define BUTTON1 2
#define BUTTON2 A4

// constants
#define NUMLEDS 10
#define PWMMAX 255 // upper limit of PWM loop, 128 levels

#define DIMDIV 1 // dimming divider

// shared variables
uint16_t led_bitmap_last, led_bitmap;  // 1 bit per LED
int16_t  led_brightness[NUMLEDS];      // array holding LED brightness for PWM
int16_t  led_brightness_inc[NUMLEDS];  // brightness incrementer for animations
int16_t  led_brightness_cyc[NUMLEDS];  // position in cycle
int i, j, pwm_now = 0;
uint8_t gCurrentPatternNumber = 0;

// Setup an array of animations, each one is the name of a function
typedef void (*SimpleAnimationList[])();
SimpleAnimationList gPatterns = { Sin8InsideOutside, RandomSparkles, Sin8slow, Sin8fast };

// helper functions
#define ARRAY_SIZE(A) (sizeof(A) / sizeof((A)[0]))

// ------------------------------------------------------------
//  Inital setup
// ------------------------------------------------------------

void setup() {
  // Setup LED outputs and turn on
  for (i = 0; i<NUMLEDS; i++) {
    pinMode(ledpins[i], OUTPUT);  // set to output
    digitalWrite(ledpins[i], 0);  // turn off
    led_brightness[i] = 0;        // set initial brightness
    led_brightness_cyc[i] = (i * 256)/NUMLEDS; // set position in cycle, e.g. sin
    led_brightness_inc[i] = 1;    // set initial incrementer for animations
  }

  // circle around and test each LED
  for (j = 0; j<NUMLEDS; j++) {
    digitalWrite(ledpins[j], 1);
    delay(100);
    digitalWrite(ledpins[j], 0);
  }
}

// ------------------------------------------------------------
//  Main loop
// ------------------------------------------------------------

void loop() {
  int i, j;

  // Each cycle through the loop increment pwm_now and update LEDs
  //  - pwm_now counts up from 0 to PWMMAX
  //  - when pwm_now reached PWMMAX the animation is updated
  //  - every 10 seconds the pattern is changed
  
  // check each LED brightness compared to pwm_now
  for (i = 0; i<NUMLEDS; i++) {
    if ((pwm_now + led_brightness[i]) == 0) {
      digitalWrite(ledpins[i], 0);
    }
    if (pwm_now < (led_brightness[i] / DIMDIV)) { // divide down the brightness for test
      digitalWrite(ledpins[i], 1);
    } else {
      digitalWrite(ledpins[i], 0);
    }
  }

  // check end of the PWM loop, update animation
  if (pwm_now++ > PWMMAX) {
    pwm_now = 0;

    // update the animation (e.g. change the brightness of each LED)
    gPatterns[gCurrentPatternNumber]();

    // change animation every few seconds
    EVERY_N_SECONDS( 10 ) { nextPattern(); }
  }
}


// ------------------------------------------------------------
//  Animations
// ------------------------------------------------------------

void Sin8slow() {
  for (i = 0; i<NUMLEDS; i++) {
    // increment position in cycle by 1, range check
    led_brightness_cyc[i]++;
    if (led_brightness_cyc[i] > 255) (led_brightness_cyc[i] -= 256);
    
    // set brightness with fast 8-bit sine function
    led_brightness[i] = sin8(led_brightness_cyc[i]);
  }
}

void Sin8fast() {
  for (i = 0; i<NUMLEDS; i++) {
    // increment position in cycle by 1, range check
    led_brightness_cyc[i] += 4;
    if (led_brightness_cyc[i] > 255) (led_brightness_cyc[i] -= 256);
    
    // set brightness with fast 8-bit sine function
    led_brightness[i] = sin8(led_brightness_cyc[i]);
  }
}

void Sin8InsideOutside() {
  for (i = 0; i<NUMLEDS; i++) {
    // increment position in cycle by 1, range check
    led_brightness_cyc[i] += 4;
    if (led_brightness_cyc[i] > 255) (led_brightness_cyc[i] -= 256);
    
    // set brightness with fast 8-bit sine function
    // even LEDs index (0,2,4,6,8) are outside
    // odd  LEDs index (1,3,5,7,9) are inside
    led_brightness[i] = sin8(led_brightness_cyc[(i&1)*5]);
  }
}

void RandomSparkles() {
  Fade2n(12, 64);
  if( random8() < 64) { 
    led_brightness[random8()%NUMLEDS] = PWMMAX;
  }
}

// next animation
void nextPattern()
{
  // add one to the current pattern number, and wrap around at the end
  gCurrentPatternNumber = (gCurrentPatternNumber + 1) % ARRAY_SIZE( gPatterns);
}

// ------------------------------------------------------------
//  Supporting functions
// ------------------------------------------------------------

// fade to black (zero brightness)
void Fade2Black(uint8_t fadespeed) {
  for (i = 0; i<NUMLEDS; i++) {
    led_brightness[i] -= fadespeed;
    if (led_brightness[i] < 0) { led_brightness[i] = 0; }
  }
}

// fade to brightness 'n'
void Fade2n(uint8_t fadespeed, uint8_t n) {
  for (i = 0; i<NUMLEDS; i++) {
    led_brightness[i] -= fadespeed;
    if (led_brightness[i] < n) { led_brightness[i] = n; }
  }
}
